/*
  Copyright 2006 CNRS
 
  Authors: Mathieu Poirier (CNRS), Florent Lamiraux (CNRS)
 
*/


/*****************************************
 INCLUDES
*******************************************/
#include <sys/time.h>

#include "KineoWorks2/kwsJointFreeFlyer.h"
#include "hppWalkPlanner/hppWalkPatGen.h"
#include "hppOpenHRP/configSwap.h"
#include "hrp2_14/hrp2_14.h"
#include "kwsIO/kwsioConfig.h"

/*****************************************
 DEFINES
*******************************************/

// Select level of verbosity at configuration.
#if DEBUG==2
#define ODEBUG2(x) std::cout << "ChppWalkPatGen:" << x << std::endl
#define ODEBUG1(x) std::cerr << "ChppWalkPatGen:" << x << std::endl
#elif DEBUG==1
#define ODEBUG2(x)
#define ODEBUG1(x) std::cerr << "ChppWalkPatGen:" << x << std::endl
#else
#define ODEBUG2(x)
#define ODEBUG1(x)
#endif

/*****************************************
 METHODS
*******************************************/


// ==========================================================================

ChppWalkPatGen::ChppWalkPatGen(double inStepLength, double inStepSpace, double inSecurityDistance) : 
  attStepLength(inStepLength), attStepSpacing(inStepSpace), attSecurityDistance(inSecurityDistance),
  attStringStream() 
{  
  double temp[40] =  HALFSITTINGPOSITION_DEG_OPENHRP ;  
  for(unsigned int i = 0 ; i < 40 ; i++) {
    attInitOpenHRPPosturePatGen[i] = temp[i]*M_PI/180.0;
  }

  // IMPORTANTE NOTE :
  // This program takes 5 arguments :
  // PATH_TO_PREVIEWCCONTRAL_PARAMS_FILE 
  // PATH_TO_VRML_FILE 
  // VRML_FILE_NAME 
  // PATH_TO_SPECIFICITIES_XML 
  // LINK_JOINT_RANK" << endl;

  string Global= WALKGENJRL_DATA "/PreviewControlParameters.ini";
  Global+= " ";
  Global+= OPENHRP_PREFIX;
  Global+="/etc/HRP2JRL/" ;
  Global+= " ";
  Global+="HRP2JRLmain.wrl" ;
  Global+= " ";
  Global+= WALKGENJRL_DATA "/HRP2Specificities.xml";
  Global+= " ";
  Global+= WALKGENJRL_DATA "/HRP2LinkJointRank.xml";
  
  
  attStringStream.str(Global);
  attWalkPatGen = new PatternGeneratorInterface(attStringStream);
  initWalkPatGen();
}

// ==========================================================================

ChppWalkPatGen::ChppWalkPatGen(double inStepLength, double inStepSpace, double inSecurityDistance, 
			       double initOpenHRPPosture[40]) :
  attStepLength(inStepLength), attStepSpacing(inStepSpace), attSecurityDistance(inSecurityDistance)
{  
  for(unsigned int i = 0 ; i < 40 ; i++) {
    attInitOpenHRPPosturePatGen[i] = initOpenHRPPosture[i];
  }

  // IMPORTANTE NOTE :
  // This program takes 5 arguments :
  // PATH_TO_PREVIEWCCONTRAL_PARAMS_FILE 
  // PATH_TO_VRML_FILE 
  // VRML_FILE_NAME 
  // PATH_TO_SPECIFICITIES_XML 
  // LINK_JOINT_RANK" << endl;

  string Global= WALKGENJRL_DATA "/PreviewControlParameters.ini";
  Global+= " ";
  Global+= OPENHRP_PREFIX;
  Global+="/etc/HRP2JRL/" ;
  Global+= " ";
  Global+="HRP2JRLmain.wrl" ;
  Global+= " ";
  Global+= WALKGENJRL_DATA "/HRP2Specificities.xml";
  Global+= " ";
  Global+= WALKGENJRL_DATA "/HRP2LinkJointRank.xml";
  
  
  attStringStream.str(Global);
  attWalkPatGen = new PatternGeneratorInterface(attStringStream);
  initWalkPatGen();
}

// ==========================================================================

void ChppWalkPatGen::reset()
{  
    //Set initial Position - half sitting position
  MAL_VECTOR_DIM(initialPosition,double,40);

  double temp[40] =  HALFSITTINGPOSITION_DEG_OPENHRP ;  
  for(unsigned int i = 0 ; i < 40 ; i++) {
    attInitOpenHRPPosturePatGen[i] = temp[i]*M_PI/180.0;
    initialPosition(i) =  attInitOpenHRPPosturePatGen[i];
  }
  attWalkPatGen->SetCurrentJointValues(initialPosition);
}

// ==========================================================================

ChppWalkPatGen::~ChppWalkPatGen()
{
  delete attWalkPatGen;
}

// ==========================================================================


ktStatus ChppWalkPatGen::initOpenHRPPosturePatGenRad(double initPos[40]) 
{
  for(unsigned int i = 0 ; i < 40 ; i++) {
    attInitOpenHRPPosturePatGen[i] = initPos[i];
  }
  return KD_OK;
}

// ==========================================================================

void ChppWalkPatGen::initWalkPatGen()
{

 const char lBuffer[9][256] =
    {":omega 0.0",
     ":stepheight 0.07",
     ":singlesupporttime 0.78",
     ":doublesupporttime 0.02",
     ":armparameters 0.5",
     ":LimitsFeasibility 0.0",
     ":ZMPShiftParameters 0.015 0.015 0.015 0.015",
     ":TimeDistributionParameters 2.0 3.7 1.7 3.0",
     ":UpperBodyMotionParameters -0.1 -1.0 0.0"
    };
  
  for(int i=0;i<9;i++) {
    std::istringstream strm(lBuffer[i]);
    attWalkPatGen->ParseCmd(strm);
  }
}

// ==========================================================================

void ChppWalkPatGen::computeStreamFromStepSequence(const ChppWalkStepSequence& inRelativeFootPrintVector)
{
  istringstream strm2 ;
  stringstream strm ;

  strm << ":stepseq " ;
  for (unsigned int count =0 ;count < inRelativeFootPrintVector.size() ; count++) {
    strm <<"\t" << inRelativeFootPrintVector[count].sx 
	 <<"\t" << inRelativeFootPrintVector[count].sy 
	 <<"\t" << inRelativeFootPrintVector[count].theta;
  }

  strm2.str(strm.str()) ;

  attWalkPatGen->ParseCmd(strm2);
}

// ==========================================================================

ktStatus ChppWalkPatGen::computeAllQFromAStepSequence(ChppDeviceShPtr hppDevice, CkwsPathShPtr kwsPathHRP2,
						      vector<CkwsConfigShPtr> &HRP2Path, vector<double> &HRP2TimeRef, 
						      vector<vector<double> > &HRP2ZMPRef, vector<vector<double> > &HRP2WaistPosition,
						      const ChppWalkStepSequence& inRelativeFootPrintVector)
{
  // collision status
  ktStatus col_state = KD_OK;

  CkwsConfigShPtr startConfig=CkwsConfig::create(hppDevice) ;
  CkwsSMLinearShPtr steeringMethod = CkwsSMLinear::create();
  bool firstIteration = true ;

  double xInit, yInit, zInit, thetaInit;

  inRelativeFootPrintVector.getInitialPosition(xInit, yInit, zInit, thetaInit);

  // Set the walking mode:
  //     - 0 : Normal mode.
  //     - 1 : Waist height variation.
  //     - 3 : Kineo works mode.
  //     - 4 : Following the upper-body motion provided by another plugin.
  // Specify the walking mode: here the default one.
  istringstream strm2(":walkmode 0");
  attWalkPatGen->ParseCmd(strm2);
  

  computeStreamFromStepSequence(inRelativeFootPrintVector) ;

  unsigned long int NbOfIt=0;
  double time=0.0 ;

  MAL_VECTOR_DIM(currentVelocity,double,46);
  MAL_VECTOR_DIM(currentConfiguration,double,46);
  MAL_VECTOR_DIM(ZMPTarget,double,3);
 


  // Main Loop of PatternGenerator Computation
  while(attWalkPatGen->RunOneStepOfTheControlLoop(currentConfiguration,
						  currentVelocity,
						  ZMPTarget)) {

    NbOfIt++;

   

    // time for file : every 0.005 secondes
    time = NbOfIt*0.005 ;

    // shift all the freeflyer position if HRP2 did not start from 0, 0
    double xPosShifted, yPosShifted, zPosShifted, rollPosShifted, pitchPosShifted, yawPosShifted ;
    xPosShifted = currentConfiguration(0)*cos(thetaInit) - currentConfiguration(1)*sin(thetaInit) + xInit ;
    yPosShifted = currentConfiguration(0)*sin(thetaInit) + currentConfiguration(1)*cos(thetaInit) + yInit ;
    zPosShifted = currentConfiguration(2)+ zInit ; 
    rollPosShifted = currentConfiguration(3) ; 
    pitchPosShifted = currentConfiguration(4) ; 
    yawPosShifted = currentConfiguration(5) + thetaInit;

    // ------------------------ Write in Data Structures ---------------//
   
    vector<double> currentZMPTarget(3) ;
    vector<double> currentWaistPosition(6) ;
    vector<double> currentOpenHrpConfig(40) ;

    HRP2TimeRef .push_back(time) ;

    currentZMPTarget[0] = ZMPTarget(0) ; currentZMPTarget[1] = ZMPTarget(1) ; currentZMPTarget[2] = ZMPTarget(2) ;
    HRP2ZMPRef.push_back(currentZMPTarget) ;

    currentWaistPosition[0] =  xPosShifted ; currentWaistPosition[1] = yPosShifted ; currentWaistPosition[2] = zPosShifted ; 
    currentWaistPosition[3] =  rollPosShifted ; currentWaistPosition[4] = pitchPosShifted ; currentWaistPosition[5] = yawPosShifted ;
    HRP2WaistPosition.push_back(currentWaistPosition) ;

    // Becareful do not insert currentConfiguration(0)->currentConfiguration(5) : 
    // because they are position and orientation 
    for( unsigned int count = 6 ; count <  MAL_VECTOR_SIZE(currentConfiguration) ; count++) {
      currentOpenHrpConfig[count-6] = currentConfiguration(count) ;
    }

    //swap OpenHRP config to Hpp config -- see package hppOpenHRP configswap.h
    CkwsConfigShPtr endConfig  =CkwsConfig::create(hppDevice) ;
    toHppConfig (currentOpenHrpConfig, currentWaistPosition , endConfig ) ;

    HRP2Path.push_back(endConfig) ;   		


    // -----------------------------------------------------------------------------//
   
    if (firstIteration) {
      /*
	At the first iteration, the initial configuration is set to
	  - the end configuration if the current path is empty
          - the end configuration of the current path otherwise
      */
      if (kwsPathHRP2->isEmpty()) {
	for (unsigned int z=0 ; z < endConfig->size() ; z++ ) {
	  startConfig->dofValue(z, endConfig->dofValue(z)) ;
	}
      }
      else {
	startConfig = kwsPathHRP2->configAtEnd();
      }

      firstIteration = false ;
    } else {

      // note : Kws can't built a Direct Path with startConfig = endConfig
      //creation of the direct path
      CkwsDPLinearShPtr DPLinear = CkwsDPLinear::create(*startConfig, *endConfig, steeringMethod) ;

      // check if the path is collising once for 2 configuration
      if(NbOfIt % 2 == 0){
	hppDevice->setCurrentConfig(*endConfig);
	if(hppDevice->collisionTest()){
	  cout<<"(self-)collision detected at t="<<NbOfIt*0.005<<endl;
	  kwsPathHRP2->clear();
	  return KD_ERROR;
	}
      }
      
      if (!DPLinear) {
	ODEBUG1(":computeAllQFromAStepSequence: failed to create direct path");
	ODEBUG1(":computeAllQFromAStepSequence: startConfig=("
		<< *startConfig << ")");
	ODEBUG1(":computeAllQFromAStepSequence: endConfig=("
		<< *endConfig << ")");
	kwsPathHRP2->clear();
	return KD_ERROR;
      }

      //add the direct path to the path
      if (kwsPathHRP2->appendDirectPath(DPLinear) == KD_ERROR) {
	ODEBUG1(":computeAllQFromAStepSequence: failed to insert direct path.");
	ODEBUG1(":computeAllQFromAStepSequence: config at end of path = ("
		<< *(kwsPathHRP2->configAtEnd()) << ")");
	ODEBUG1(":computeAllQFromAStepSequence: config at start of direct path = ("
		<< *(DPLinear->configAtStart()) << ")");
	kwsPathHRP2->clear();
	return KD_ERROR;
      }

      for (unsigned int z=0 ; z < endConfig->size() ; z++ ) {
	startConfig->dofValue(z, endConfig->dofValue(z)) ;
      }
    }
  }

  ODEBUG1("");
  ODEBUG1(" number of iterations " << NbOfIt);

  return col_state;
}

